import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Tuple

from aiofile import async_open

from .abc import PlaytimeDate, TopActivity, TrendingActivity
from .constants import API
from .utils import HUMANIZE_DAYS, HUMNANIZE_HOURS, humanize_duration, icon_to_bytes

if TYPE_CHECKING:
    from io import BytesIO

__all__: Tuple[str, ...] = (
    "User",
    "UserStats",
)


class User:
    """
    Class Interface Containing User Attributes

    Attributes
    ----------
    id: :class:`int`
        PresenceDB Internal User ID
    dId: :class:`int`
        User Discord ID
    name: :class:`str`
        User's Name
    discriminator: :class:`str`
        User's Discord Discriminator
    inactive: :class:`bool`
        If the User is marked as Inactive
    added: :class:`str`
        Date User's Account Was Added
    color: :class:`str`
        Color Of User's Account
    avatar: :class:`str`
        Avatar ID of User
    plus: :class:`bool`
        If The User Has Subscribed To PresenceDB Plus
    timestamp: :class:`str`
        Timestamp of Current Activity
    current_activities: List[CurrentActivity]
        List of Current User Activities
    stats: UserStats
        Object Representing User Statistics
    tag: :class:`str`
        Combination of `self.name` + `self.discriminator`
    """

    __slots__: Tuple[str, ...] = (
        "id",
        "dId",
        "name",
        "discriminator",
        "inactive",
        "private",
        "added",
        "color",
        "avatar",
        "plus",
        "timestamp",
        "current_activities",
        "stats",
        "tag",
    )

    def __init__(self, data: Dict, stats: Dict, should_format: bool) -> None:
        self.id: int = data["id"]
        self.dId: int = data["dId"]
        self.name: str = data["name"]
        self.discriminator: str = data["discriminator"]
        self.inactive: bool = data["inactive"]
        self.private: bool = data["private"]
        self.added: str = data["added"]
        self.color: str = data["color"]
        self.avatar: str = data["avatar"]
        self.plus: bool = data["plus"]
        self.timestamp: str | None = (
            data["tracker"]["timestamp"] if data["tracker"] else None
        )
        self.current_activities: CurrentActivity | None = (
            [CurrentActivity(**activity) for activity in data["tracker"]["activities"]]
            if data["tracker"]
            else None
        )
        self.stats: UserStats = UserStats(stats, should_format)
        self.tag: str = self.name + f"#{self.discriminator}"

        def __repr__(self):
            return self.tag

        def __eq__(self, other):
            return self.id == other.id

        def __hash__(self):
            return self.id

    @property
    def icon_url(self) -> str:
        """Get Avatar URL of User

        Returns
        -------
        str
            Avatar URL
        """
        return f"{API.ICON_BASE}/{self.dId}/{self.icon}"

    async def save_avatar(self, path: os.PathLike) -> None:
        """Saves Current Activity To File

        Parameters
        ----------
        path : os.PathLike
            Path To Save File Too
        """
        bytes: BytesIO = await icon_to_bytes(self.icon_url)
        async with async_open(path, "wb") as file:
            await file.write(bytes.read())


class UserStats:
    """
    Class Interface Representing Stats of a User

    Attributes
    ----------
    total_duration: :class:`str`
        Total duration of activity recorded
    trending_duration: :class:`str`
        Trending Duration of Activities
    top_users: List[TopUser]
        List of Top Users For The Activity
    playtime_dates: List[PlaytimeDate]
        List of Playtime Dates For Activity
    top_activities: List[TopActivity]
        List of Top Activities For User
    trending_activities: List[TrendingActivity]
        List of Trending Activities For User
    avatar_history: List[AvatarHistory]
        List of Avatars Users Recorded
    """

    __slots__: Tuple[str, ...] = (
        "total_duration",
        "trending_duration",
        "playtime_dates",
        "top_activities",
        "trending_activities",
        "avatar_history",
    )

    def __init__(self, stats: Dict, should_format) -> None:
        self.total_duration: str = (
            stats["totalDuration"]
            if not should_format
            else humanize_duration(stats["totalDuration"], HUMANIZE_DAYS)
        )
        self.trending_duration: str = (
            stats["trendingDuration"]
            if not should_format
            else humanize_duration(stats["trendingDuration"], HUMNANIZE_HOURS)
        )
        self.playtime_dates: List[PlaytimeDate] = [
            PlaytimeDate(**playtime_date) for playtime_date in stats["playtimeDates"]
        ]
        self.top_activities: List[TopActivity] = [
            TopActivity(**activity) for activity in stats["topActivities"]
        ]
        self.trending_activities: List[TrendingActivity] = [
            TrendingActivity(**activity) for activity in stats["trendingActivities"]
        ]
        self.avatar_history: List[AvatarHistory] = [
            AvatarHistory(**avatar) for avatar in stats["avatarHistory"]
        ]


@dataclass
class CurrentActivity:
    """
    Class Representing A Current Activity

    Attributes
    ----------
    name: :class:`str`
        Name of Activity
    id: :class:`int`
        ID of Activity
    """

    name: str
    id: int


@dataclass
class AvatarHistory:
    """
    Class Representing A User's Avatar History

    Attributes
    ----------
    id: :class:`int`
        ID of Avatar
    dId: :class:`int`
        Discord ID Pertaining to Avatar
    avatar: :class:`str`
        Avatar URL
    added: :class:`str`
        Date Avatar Was Added
    hidden: :class:`bool`
        Whether Avatar Is Hidden From Site
    """

    id: int
    dUserId: int
    avatar: str
    added: str
    hidden: bool

    def __post_init__(self):
        self.avatar = f"{API.AVATAR_BASE}/{self.dUserId}/{self.avatar}"



import aiohttp
import pytest

from presencedb.client import Client
from presencedb.enums import ActivityID


@pytest.mark.asyncio
async def test_fetch_activity() -> None:
    session = aiohttp.ClientSession()
    client = Client(session=session)

    activity = await client.get_activity(ActivityID.VSCODE)
    await client.cleanup()

    assert activity.color == "#FFFFFF"
    assert activity.name == "Visual Studio Code"


@pytest.mark.asyncio
async def test_fetch_top_activities() -> None:
    session = aiohttp.ClientSession()
    client = Client(session=session)

    top_activities = await client.get_top_activities()
    await client.cleanup()

    assert top_activities[0].name == "Spotify"


@pytest.mark.asyncio
async def test_fetch_trending_activities() -> None:
    session = aiohttp.ClientSession()
    client = Client(session=session)

    trending_activities = await client.get_trending_activities()
    await client.cleanup()

    assert trending_activities[0].name == "Spotify"



from dataclasses import dataclass
from typing import Tuple

from .constants import API

__all__: Tuple[str, ...] = (
    "TopUser",
    "TopActivity",
    "TrendingActivity",
    "TopActivity",
)


@dataclass
class TopUser:
    """
    Class Representing A Top User

    Attributes
    ----------
    name: :class:`str`
        Name of User
    avatar: :class:`str`
        Avatar URL
    discriminator: :class:`str`
        User Discriminator
    dId: :class:`int`
        Discord ID of User
    duration: :class:`int`
        Duration User Has on Activity
    """

    name: str
    avatar: str
    discriminator: str
    dId: int
    duration: int

    def __post_init__(self):
        self.avatar = f"{API.AVATAR_BASE}/{self.dId}/{self.avatar}"


@dataclass
class TopActivity:
    """
    Class Representing A Top Activity

    Attributes
    ----------
    name: :class:`str`
        Name of Activity
    dId: :class:`int`
        ID of Activity
    icon: :class:`str`
        Activity Icon URL
    duration: :class:`int`
        Duration Activity Was Played For
    """

    name: str
    dId: int
    icon: str
    duration: int

    def __post_init__(self):
        self.icon = f"{API.ICON_BASE}/{self.dId}/{self.icon}"


@dataclass
class TrendingActivity:
    """
    Class Representing A Trending Activity

    Attributes
    ----------
    name: :class:`str`
        Name of Activity
    dId: class:`int`
        ID of Activity
    icon: :class:`str`
        Activity Icon URL
    duration: :class:`int`
        Duration Activity Has Been Played For
    """

    name: str
    dId: int
    icon: str
    duration: int

    def __post_init__(self):
        self.icon = f"{API.ICON_BASE}/{self.dId}/{self.icon}"


@dataclass
class PlaytimeDate:
    """
    Class Representing A Playtime Date

    Attributes
    ----------
    date: :class:`str`
        Date Activity Was Played On
    seconds: :class:`int`
        Duration Activity Was Played For In Seconds
    """

    date: str
    seconds: int



import aiohttp
import pytest

from presencedb.client import Client

USER_ACCOUNT_ID: str = "433026067050266634"


@pytest.mark.asyncio
async def test_fetch_user() -> None:
    session = aiohttp.ClientSession()
    client = Client(session=session)

    user = await client.get_user(USER_ACCOUNT_ID)
    await client.cleanup()

    assert user.dId == USER_ACCOUNT_ID
    assert user.name == "ghoul"



from typing import Tuple

from .abc import *
from .activity import *
from .client import Client
from .enums import ActivityID
from .errors import *
from .user import *
from .utils import humanize_duration, icon_to_bytes

__all__: Tuple[str, ...] = (
    "Client",
    "ActivityID",
    "icon_to_bytes",
    "humanize_duration",
)

__version__ = "2.0.0"
__author__ = "Ghoul"
__license__ = "MIT"



import datetime
import io
from typing import Final, List, Optional, Tuple, Union

import aiohttp
import humanize

__all__: Tuple[str, ...] = (
    "icon_to_bytes",
    "humanize_duration",
)

HUMNANIZE_HOURS: Final[str] = "hours"
HUMANIZE_DAYS: Final[str] = "days"


async def icon_to_bytes(icon: str) -> io.BytesIO:
    """Converts Icon URL To Bytes

    Parameters
    ----------
    icon : :class:`str`
        Icon URL

    Returns
    -------
    io.BytesIO
        Bytes Like Object Of Icon
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(icon) as response:
            buffer = io.BytesIO(await response.read())
            return buffer


def humanize_duration(
    number: int, type: Optional[Union[HUMNANIZE_HOURS, HUMANIZE_DAYS]] = HUMNANIZE_HOURS
) -> str:
    """Generates a Human Readable Duration

    Parameters
    ----------
    number : int
        Duration To Format
    type : Optional[Union[HUMNANIZE_HOURS, HUMANIZE_DAYS]], optional
        If The Output Should Be Days or Hours, defaults to HUMNANIZE_HOURS

    Returns
    -------
    str
        Humanized Duration
    """
    if type == HUMNANIZE_HOURS:
        suppress: List[str] = [
            "seconds",
            "minutes",
            "seconds",
            "days",
            "years",
            "months",
        ]
    elif type == HUMANIZE_DAYS:
        suppress: List[str] = [
            "seconds",
            "minutes",
            "seconds",
            "hours",
            "years",
            "months",
        ]

    duration: str = humanize.precisedelta(
        datetime.timedelta(seconds=number), suppress=suppress, format="%0.1f"
    )
    return duration



from typing import Dict, List, Tuple

from .abc import PlaytimeDate, TopUser
from .constants import API
from .utils import HUMANIZE_DAYS, HUMNANIZE_HOURS, humanize_duration

__all__: Tuple[str, ...] = (
    "Activity",
    "ActivityStats",
)


class Activity:
    """
    Class Interface Representing An Activity

    Attributes
    ----------
    name: :class:`str`
        Name of Activity
    dId: :class:`int`
        ID of Activity
    added: :class:`str`
        Date Activity Was Added
    icon: :class:`str`
        Avatar Icon ID
    color: :class:`str`
        Color of Activity
    stats: ActivityStats
        Stats of Activity
    """

    __slots__: Tuple[str, ...] = (
        "name",
        "dId",
        "added",
        "icon",
        "color",
        "stats",
    )

    def __init__(self, data: Dict, stats: Dict, should_format: bool) -> None:
        self.name: str = data["name"]
        self.dId: int = data["dId"]
        self.added: str = data["added"]
        self.icon: str = data["icon"]
        self.color: str = data["color"]
        self.stats: ActivityStats = ActivityStats(stats, should_format)

        def __repr__(self):
            return self.name

    @property
    def icon_url(self) -> str:
        """Get Icon URL of Activity

        Returns
        -------
        str
            Icon URL
        """
        return f"{API.ICON_BASE}/{self.dId}/{self.icon}"


class ActivityStats:
    """
    Class Representing Stats of an Activity

    Attributes
    ----------
    total_duration: :class:`str`
        Total duration of activity recorded
    trending_duration: :class:`str`
        Trending Duration of Activities
    top_users: List[TopUser]
        List of Top Users For The Activity
    playtime_dates: List[PlaytimeDate]
        List of Playtime Dates For Activity
    """

    def __init__(self, stats: Dict, should_format: bool) -> None:
        self.total_duration: str = (
            stats["totalDuration"]
            if not should_format
            else humanize_duration(stats["totalDuration"], HUMANIZE_DAYS)
        )
        self.trending_duration: str = (
            stats["trendingDuration"]
            if not should_format
            else humanize_duration(stats["trendingDuration"], HUMNANIZE_HOURS)
        )
        self.top_users: List[TopUser] = [
            TopUser(**top_user) for top_user in stats["topUsers"]
        ]
        self.playtime_dates: List[PlaytimeDate] = [
            PlaytimeDate(**playtime_date) for playtime_date in stats["playtimeDates"]
        ]



from typing import Dict, Final, Tuple, final

from yarl import URL

__all__: Tuple[str, ...] = ("API",)


@final
class API:
    """Class Representing Constants For API"""

    BASE_URL: Final[URL] = URL("https://api.presencedb.com")
    ICON_BASE: Final[URL] = URL("https://s3.vasc.dev/presencedb/app-icons")
    AVATAR_BASE: Final[URL] = URL("https://s3.vasc.dev/presencedb/avatars")
    HEADERS: Final[Dict[str, str]] = {
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36 OPR/82.0.4227.50",
        "content-type": "application/json",
    }



from typing import Tuple

__all__: Tuple[str, ...] = (
    "PresenceDBException",
    "UserNotFound",
    "ActivityNotFound",
    "HTTPError",
    "ActivitiesNotFound",
)


class PresenceDBException(Exception):
    """Generic API Exception For PresenceDB

    Attributes
    -----------
    detail: :class:`str`
        The details of this request
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    def __init__(self, detail: str, status_code: int) -> None:
        self.detail: str = detail
        self.status_code: int = status_code
        super().__init__(detail, status_code)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} detail={self.detail} status_code={self.status_code}>"


class UserNotFound(PresenceDBException):
    """Raised When User Could Not Be Found

    Attributes
    -----------
    detail: :class:`str`
        The details of this request
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    def __init__(self, detail: str, status_code: int) -> None:
        super().__init__(detail="User Could Not Be Found", status_code=404)


class ActivityNotFound(PresenceDBException):
    """Raised When User Could Not Be Found

    Attributes
    -----------
    detail: :class:`str`
        The details of this request
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    def __init__(self, detail: str, status_code: int) -> None:
        super().__init__(detail="Activity Could Not Be Found", status_code=404)


class ActivitiesNotFound(PresenceDBException):
    """Raised When User Could Not Be Found

    Attributes
    -----------
    detail: :class:`str`
        The details of this request
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    def __init__(self, detail: str, status_code: int) -> None:
        super().__init__(detail="Activities Could Not Be Found", status_code=404)


class HTTPError(PresenceDBException):
    """Raised When User Could Not Be Found

    Attributes
    -----------
    detail: :class:`str`
        The details of this request
    status_code: :class:`int`
        The HTTP status code for this request.
    """

    def __init__(self, detail: str, status_code: int) -> None:
        super().__init__(
            detail="Activities Could Not Be Found", status_code=status_code
        )



from distutils.core import setup

setup(
    name="presencedb",
    packages=["presencedb"],
    version="2.0.0",
    license="MIT",
    description="API wrapper for the presencedb (undocumented) api",
    author="Ghoul",
    url="https://github.com/xFGhoul/py-presencedb",
    keywords=["discord", "api", "presencedb", "discord-api", "discord bot"],
    install_requires=["aiohttp", "humanize"],
    extra_requires={"docs": ["mkdocs-material"], "tests": ["pytest"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
)



import logging
from typing import Dict, List, Optional, Tuple, Union

import aiohttp
from aiohttp import ClientSession

from .abc import TopActivity, TrendingActivity
from .activity import Activity
from .constants import API
from .enums import ActivityID
from .errors import ActivitiesNotFound, ActivityNotFound, UserNotFound
from .user import User

__all__: Tuple[str, ...] = ("Client",)


class Client:
    """Client Interface For Interacting With PresenceDB API

    Parameters
    -----------
    session: Optional[aiohttp.ClientSession]
        Client Session Used For HTTP Requests
    """

    __slots__ = (
        "session",
        "logger",
    )

    def __init__(self, session: Optional[ClientSession] = None) -> None:
        self.logger = logging.getLogger(__name__)
        if session is None:
            self.session = aiohttp.ClientSession(headers=API.HEADERS)
        else:
            self.session = session

    async def _make_request(self, target: str) -> Dict:
        """

        Sends An HTTP /GET Request To PresenceDB API

        Parameters
        ----------
        target: :class:`str`
            API Endpoint

        Returns
        -------
        Dict
            JSON Response From API

        Raises
        ------
        HTTPError
            If The Request Failed
        """
        async with self.session.get(API.BASE_URL / target) as response:
            self.logger.debug(f"Sending HTTP /GET Request: {API.BASE_URL}/{target}")
            json = await response.json()
            return json

    async def cleanup(self) -> None:
        """Closes The Current Client Session

        Raises
        ------
        RuntimeError
            If The Session Was Already Closed
        """
        if self.session.closed:
            raise RuntimeError("Client Session Already Closed.")
        self.logger.debug("Closing Session")
        await self.session.close()
        self.logger.debug("Session Closed")

    async def get_user(self, user_id: int, format: Optional[bool] = False) -> User:
        """Get A User's Profile

        Parameters
        ----------
        user_id : :class:`int`
            Discord ID of User
        format : Optional[:class:`bool`]
            If To Format User Data, by default False

        Returns
        -------
        User
            Object Response Of User

        Raises
        ------
        UserNotFound
            If The User Was Not Found
        """
        data = await self._make_request(f"user/{user_id}")
        stats = await self._make_request(f"user/{user_id}/stats")
        if data and stats is not None:
            return User(data, stats, format)
        else:
            raise UserNotFound

    async def get_activity(
        self, activity_id: Union[int, ActivityID], format: Optional[bool] = False
    ) -> Activity:
        """Get An Activity

        Parameters
        ----------
        activity_id : Union[:class:`int`, ActivityID]
            ID of Activity
        format : Optional[:class:`bool`]
            If Duration Values Should Be Formatted, by default False

        Returns
        -------
        Activity
            Object Of JSON Response

        Raises
        ------
        ActivityNotFound
            If The Activity Could Not Be Found
        """
        data = await self._make_request(f"activity/{activity_id}")
        stats = await self._make_request(f"activity/{activity_id}/stats")
        if data and stats is not None:
            return Activity(data, stats, format)
        else:
            raise ActivityNotFound

    async def get_top_activities(self) -> List[TopActivity]:
        """Returns Top Activities

        Returns
        -------
        List[TopActivity]
            List Of Top Activities

        Raises
        ------
        ActivitiesNotFound
            Top Activities Could Not Be Fetched
        """
        data = await self._make_request("activities/top")
        if data:
            top_activities = [TopActivity(**activity) for activity in data]
            return top_activities
        else:
            raise ActivitiesNotFound

    async def get_trending_activities(self) -> List[TrendingActivity]:
        """Returns Current Trending Activities

        Returns
        -------
        List[TrendingActivity]
            List Of Trending Activities
        Raises
        ------
        ActivitiesNotFound
            Trending Activities Could Not Be Fetched
        """
        data = await self._make_request("activities/trending")
        if data:
            trending_activities = [TrendingActivity(**activity) for activity in data]
            return trending_activities
        else:
            raise ActivitiesNotFound
